.details-popover contents {
  padding: 12px;
}

.details-popover .information {
  border-spacing: 6px;
}

.details-popover .actions {
  margin-top: 6px;
  border-spacing: 12px;
}

.details-popover .actions button {
  padding: 12px;
}
